//
// Copyright 2003 (c) Kevin Atkinson under the GNU GPL license version
// 2.0.  You should have received a copy of the GPL license along with
// this program if you did not you can find it at http://www.gnu.org/.
//

#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>

#include "avisynth_c.h"

typedef unsigned char byte;

AVS_VideoFrame * ibob_get_frame(AVS_FilterInfo *, int num);
void ibob_free_filter(AVS_FilterInfo *);

typedef struct Filter
{
  int cur;
  AVS_VideoFrame * cache[2];
  bool used[2];
  bool even_first;
} Filter;

void filter_reset(Filter * f) {
  if (f->cache[0]) avs_release_frame(f->cache[0]);
  if (f->cache[1]) avs_release_frame(f->cache[1]);
  f->cur = -1;
  f->cache[0] = 0; f->cache[1] = 0;
  f->used[0] = false; f->used[1] = false;
}

AVS_Value ibob(AVS_ScriptEnvironment * env, AVS_Value args, void * d)
{
  AVS_Value ret = avs_void;
  AVS_FilterInfo * fi = 0;

  AVS_Clip * clip = avs_new_c_filter(env, &fi, avs_array_elt(args,0), 1);
  fi->user_data = calloc(sizeof(Filter),1);
  Filter * f = (Filter *)fi->user_data;
  f->cur = -1;
  f->even_first = !avs_is_bff(&fi->vi);
  fi->get_frame = ibob_get_frame;
  fi->free_filter = ibob_free_filter;

  avs_clear_property(&fi->vi, AVS_IT_FIELDBASED);
  avs_clear_property(&fi->vi, AVS_IT_TFF);
  avs_clear_property(&fi->vi, AVS_IT_BFF);
  fi->vi.fps_numerator *= 2;
  fi->vi.num_frames *= 2;

  ret = avs_new_value_clip(clip);
  avs_release_clip(clip);
  return ret;
}

const char * avisynth_c_plugin_init(AVS_ScriptEnvironment * env)
{
  avs_add_function(env, "IBob", "c", ibob, 0);
  return "";
}


void bob_both(const byte * s, int s_pitch,
	      byte * e, byte * o, int o_pitch,
	      int row_size, int height)
{
  for (int x = 0; x < row_size; ++x) {
    o[x]           = s[x + s_pitch];
    o[x + o_pitch] = s[x + s_pitch];
  }
  s += 2*s_pitch;
  e += 2*o_pitch;
  o += 3*o_pitch;
  for (int y = 0; y != height/2-1; ++y) {
    for (int x = 0; x < row_size; ++x) {
      e[x - 2*o_pitch] = s[x - 2*s_pitch];
      e[x - o_pitch]   = (s[x - 2*s_pitch] + s[x] + 1)/2;
      o[x - o_pitch]   = (s[x - s_pitch] + s[x + s_pitch] + 1)/2;
      o[x]             = s[x + s_pitch];
    }
    s += 2*s_pitch;
    e += 2*o_pitch;
    o += 2*o_pitch;
  }
  for (int x = 0; x < row_size; ++x) {
    e[x - 2*o_pitch] = s[x - 2*s_pitch];
    e[x - o_pitch] = s[x - 2*s_pitch];
  }
}

void bob_both_inplace(byte * e, int e_pitch, byte * o, int o_pitch,
		      int row_size, int height)
{
  for (int x = 0; x < row_size; ++x) {
    o[x]           = e[x + e_pitch];
    o[x + o_pitch] = e[x + e_pitch];
  }
  e += 2*e_pitch;
  o += 3*o_pitch;
  for (int y = 0; y != height/2-1; ++y) {
    for (int x = 0; x < row_size; ++x) {
      o[x - o_pitch] = (e[x - e_pitch] + e[x + e_pitch] + 1)/2;
      o[x]           = e[x + e_pitch];
      e[x - e_pitch] = (e[x - 2*e_pitch] + e[x] + 1)/2;
    }
    e += 2*e_pitch;
    o += 2*o_pitch;
  }
  for (int x = 0; x < row_size; ++x) {
    e[x - e_pitch] = e[x - 2*e_pitch];
  }
}

void bob_even(const byte * s, int s_pitch, byte * d, int d_pitch,
	      int row_size, int height)
{
  height /= 2;
  s_pitch *= 2;
  byte * d0 = d;
  byte * d1 = d0 + d_pitch;
  d_pitch *= 2;
  const byte * s0 = s; 
  const byte * s1 = s0 + s_pitch;
  for (int y = 0; y != height - 1; ++y) {
    for (int x = 0; x != row_size; ++x) {
      d0[x] = s0[x];
      d1[x] = (s0[x] + s1[x] + 1)/2;
    }
    d0 += d_pitch; d1 += d_pitch;
    s0 = s1; s1 += s_pitch;
  }
  for (int x = 0; x != row_size; ++x) {
    d0[x] = s0[x];
    d1[x] = s0[x];
  }
}

void bob_odd(const byte * s, int s_pitch, byte * d, int d_pitch,
	     int row_size, int height)
{
  height /= 2;
  s += s_pitch;
  s_pitch *= 2;
  byte * d0 = d;
  byte * d1 = d0 + d_pitch;
  d_pitch *= 2;
  const byte * s0 = 0; 
  const byte * s1 = s;
  for (int x = 0; x != row_size; ++x) {
    d0[x] = s1[x];
    d1[x] = s1[x];
  }
  for (int y = 1; y != height; ++y) {
    d0 += d_pitch; d1 += d_pitch;
    s0 = s1; s1 += s_pitch;
    for (int x = 0; x != row_size; ++x) {
      d0[x] = (s0[x] + s1[x] + 1)/2;
      d1[x] = s1[x];
    }
  }
}

void bob_even_inplace(byte * d, int d_pitch,
		      int row_size, int height)
{
  height /= 2;
  byte * d0 = d;
  byte * d1 = d0 + d_pitch;
  byte * d2 = d1 + d_pitch;
  d_pitch *= 2;
  for (int y = 0; y != height - 1; ++y) {
    for (int x = 0; x != row_size; ++x) {
      d1[x] = (d0[x] + d2[x] + 1)/2;
    }
    d0 = d2; d1 += d_pitch; d2 += d_pitch;
  }
  for (int x = 0; x != row_size; ++x) {
    d1[x] = d0[x];
  }
}

void bob_odd_inplace(byte * d, int d_pitch,
		     int row_size, int height)
{
  height /= 2;
  byte * d0 = 0;
  byte * d1 = d;
  byte * d2 = d1 + d_pitch;
  d_pitch *= 2;
  for (int x = 0; x != row_size; ++x) {
    d1[x] = d2[x];
  } 
  for (int y = 1; y != height; ++y) {
    d0 = d2; d1 += d_pitch; d2 += d_pitch;
    for (int x = 0; x != row_size; ++x) {
      d1[x] = (d0[x] + d2[x] + 1)/2;
    }
  }
}

typedef struct Bob {
  void (* bob)(const byte *, int s_pitch,
	       byte * d, int d_pitch,
	       int row_size, int height);
  void (* bob_inplace)(byte * d, int d_pitch,
		       int row_size, int height);
} Bob;

static const Bob bob_witch[2] = {{bob_even, bob_even_inplace}, 
				 {bob_odd, bob_odd_inplace}};

void bob_chroma_both(const byte * s, int s_pitch,
		     byte * e, byte * o, int o_pitch,
		     int row_size, int height)
{
  for (int y = 0; y != height/2; ++y) {
    for (int x = 0; x != row_size; ++x) {
      e[x] = s[x];
      e[x + o_pitch] = s[x];
      o[x] = s[x + o_pitch];
      o[x + o_pitch] = e[x + o_pitch];
    }
    s += s_pitch*2;
    e += s_pitch*2;
    o += o_pitch*2;
  }
}

void bob_chroma_both_inplace(byte * e, int e_pitch, 
			     byte * o, int o_pitch,
			     int row_size, int height)
{
  for (int y = 0; y != height/2; ++y) {
    for (int x = 0; x != row_size; ++x) {
      o[x] = e[x + e_pitch];
      o[x + o_pitch] = e[x + e_pitch];
      e[x + e_pitch] = e[x];
    }
    e += e_pitch*2;
    o += o_pitch*2;
  }
}

void bob_chroma(const byte * s, int s_pitch, byte * d, int d_pitch,
		int row_size, int height, bool use_even)
{
  if (!use_even) s += s_pitch;
  s_pitch *= 2;
  height /= 2;
  byte * d0 = d;
  byte * d1 = d0 + d_pitch;
  d_pitch *= 2;
  for (int y = 0; y != height; ++y) {
    for (int x = 0; x != row_size; ++x) {
      d0[x] = s[x];
      d1[x] = s[x];
    }
    d0 += d_pitch; d1 += d_pitch;
    s += s_pitch;
  }
}

void bob_chroma_inplace(byte * d, int pitch,
			int row_size, int height, bool use_even)
{
  byte * s;
  if (use_even) {
    s = d;
    d = d + pitch;
  } else {
    s = d + pitch;
  }
  pitch *= 2;
  height /= 2;
  for (int y = 0; y != height; ++y) {
    for (int x = 0; x != row_size; ++x) {
      d[x] = s[x];
    }
    d += pitch;
    s += pitch;
  }
}

#define LOG(str)

AVS_VideoFrame * ibob_get_frame(AVS_FilterInfo * fi, int num)
{
  enum What {both, oneof, r_even, r_odd} what;
  Filter * f = (Filter *)fi->user_data;
  if (f->cur != num / 2) {
    if (f->used[0] && f->used[1]) what = both;
    else                          what = oneof;
    filter_reset(f);
    f->cur = num / 2;
  } else {
    AVS_VideoFrame * c = f->cache[num % 2];
    if (c) {
      LOG("cached");
      f->cache[num % 2] = 0;
      f->used[num % 2] = true;
      return c;
    } else {
      what = oneof;
    }
  }
  f->used[num % 2] = true;

  if (what == oneof) {
    if      (num % 2 == 0 && f->even_first) what = r_even;
    else if (num % 2 == 0)                  what = r_odd;
    else if (num % 2 != 0 && f->even_first) what = r_odd;
    else                                    what = r_even;
  }

  AVS_VideoFrame * src = avs_get_frame(fi->child, num / 2);

  if (what == both) {

    AVS_VideoFrame * even;
    AVS_VideoFrame * odd  = avs_new_frame(fi->env, &fi->vi);
  
    if (avs_is_writable(src)) {
      LOG("both inplace");
      even = src;
      bob_both_inplace(avs_get_write_ptr(even), avs_get_pitch(even),
		       avs_get_write_ptr(odd), avs_get_pitch(odd),
		       avs_get_row_size(even), avs_get_height(even));

      if (avs_is_yv12(&fi->vi)) {
	bob_chroma_both_inplace(avs_get_write_ptr_p(even, AVS_PLANAR_U), 
				avs_get_pitch_p(even, AVS_PLANAR_U),
				avs_get_write_ptr_p(odd, AVS_PLANAR_U), 
				avs_get_pitch_p(odd, AVS_PLANAR_U),
				avs_get_row_size_p(even, AVS_PLANAR_U), 
				avs_get_height_p(even, AVS_PLANAR_U));
	bob_chroma_both_inplace(avs_get_write_ptr_p(even, AVS_PLANAR_V), 
				avs_get_pitch_p(even, AVS_PLANAR_V),
				avs_get_write_ptr_p(odd, AVS_PLANAR_V), 
				avs_get_pitch_p(odd, AVS_PLANAR_V),
				avs_get_row_size_p(even, AVS_PLANAR_V), 
				avs_get_height_p(even, AVS_PLANAR_V));
      }
    } else {
      even = avs_new_frame(fi->env, &fi->vi);
      LOG("both");
      assert(avs_get_pitch(even) == avs_get_pitch(odd));
      bob_both(avs_get_read_ptr(src), avs_get_pitch(src),
	       avs_get_write_ptr(even), 
	       avs_get_write_ptr(odd), avs_get_pitch(odd),
	       avs_get_row_size(src), avs_get_height(src));
      
      if (avs_is_yv12(&fi->vi)) {
	bob_chroma_both(avs_get_read_ptr_p(src, AVS_PLANAR_U), 
			avs_get_pitch_p(src, AVS_PLANAR_U),
			avs_get_write_ptr_p(even, AVS_PLANAR_U), 
			avs_get_write_ptr_p(odd, AVS_PLANAR_U), 
			avs_get_pitch_p(odd, AVS_PLANAR_U),
			avs_get_row_size_p(src, AVS_PLANAR_U), 
			avs_get_height_p(src, AVS_PLANAR_U));
	bob_chroma_both(avs_get_read_ptr_p(src, AVS_PLANAR_V), 
			avs_get_pitch_p(src, AVS_PLANAR_V),
			avs_get_write_ptr_p(even, AVS_PLANAR_V), 
			avs_get_write_ptr_p(odd, AVS_PLANAR_V), 
			avs_get_pitch_p(odd, AVS_PLANAR_V),
			avs_get_row_size_p(src, AVS_PLANAR_V), 
			avs_get_height_p(src, AVS_PLANAR_V));
      }
      avs_release_frame(src);
    }
    
    if (f->even_first) {f->cache[0] = even; f->cache[1] = odd;}
    else               {f->cache[1] = even; f->cache[0] = odd;}

    AVS_VideoFrame * c = f->cache[num % 2];
    f->cache[num % 2] = 0;
    return c;
  
  } else {

    AVS_VideoFrame * dest;
    bool witch = what == r_even ? 0 : 1;

    if (avs_is_writable(src)) {
      LOG("oneof inplace");
      dest = src;
      bob_witch[witch].bob_inplace(avs_get_write_ptr(src), avs_get_pitch(src),
				   avs_get_row_size(src), avs_get_height(src));
      if (avs_is_yv12(&fi->vi)) {
	bob_chroma_inplace(avs_get_write_ptr_p(src, AVS_PLANAR_U), 
			   avs_get_pitch_p(src, AVS_PLANAR_U),
			   avs_get_row_size_p(src, AVS_PLANAR_U), 
			   avs_get_height_p(src, AVS_PLANAR_U),
			   witch);
	bob_chroma_inplace(avs_get_write_ptr_p(src, AVS_PLANAR_V), 
			   avs_get_pitch_p(src, AVS_PLANAR_V),
			   avs_get_row_size_p(src, AVS_PLANAR_V), 
			   avs_get_height_p(src, AVS_PLANAR_V),
			   witch);
      }
    } else {
      LOG("oneof");
      dest = avs_new_frame(fi->env, &fi->vi);
      bob_witch[witch].bob(avs_get_read_ptr(src), avs_get_pitch(src),
			   avs_get_write_ptr(dest), avs_get_pitch(dest),
			   avs_get_row_size(src), avs_get_height(src));
      if (avs_is_yv12(&fi->vi)) {
	bob_chroma(avs_get_read_ptr_p(src, AVS_PLANAR_U), 
		   avs_get_pitch_p(src, AVS_PLANAR_U),
		   avs_get_write_ptr_p(dest, AVS_PLANAR_U), 
		   avs_get_pitch_p(dest, AVS_PLANAR_U),
		   avs_get_row_size_p(src, AVS_PLANAR_U), 
		   avs_get_height_p(src, AVS_PLANAR_U),
		   witch);
	bob_chroma(avs_get_read_ptr_p(src, AVS_PLANAR_V), 
		   avs_get_pitch_p(src, AVS_PLANAR_V),
		   avs_get_write_ptr_p(dest, AVS_PLANAR_V), 
		   avs_get_pitch_p(dest, AVS_PLANAR_V),
		   avs_get_row_size_p(src, AVS_PLANAR_V), 
		   avs_get_height_p(src, AVS_PLANAR_V),
		   witch);
      }
      avs_release_frame(src);
    }
    return dest;
  }
}

void ibob_free_filter(AVS_FilterInfo * fi)
{
  if (fi->user_data) free(fi->user_data);
  fi->user_data = 0;
}
