//
// Copyright 2003 (c) Kevin Atkinson under the GNU GPL license version
// 2.0.  You should have received a copy of the GPL license along with
// this program if you did not you can find it at http://www.gnu.org/.
//

#include <stdio.h>
#include <fcntl.h>
#include <windows.h>
#include <io.h>

#include "avisynth_c.h"

static FILE * gclog = 0;
void alloc_clog()
{
  if (gclog) return;
  AllocConsole();
  HANDLE h = GetStdHandle( STD_OUTPUT_HANDLE );
  int fd = _open_osfhandle( (long) h, _O_TEXT );
  gclog = fdopen(fd, "w");
}

typedef unsigned char byte;

AVS_VideoFrame * ibob_get_frame(AVS_FilterInfo *, int num);
void ibob_free_filter(AVS_FilterInfo *);

#define ERR(str) do {ret = avs_new_value_error(str); goto error;} while (0)

AVS_Value ibob(AVS_ScriptEnvironment * env, AVS_Value args, void * )
{
  AVS_Value ret = avs_void, child_v = avs_void;
  AVS_FilterInfo * fi = 0;

  AVS_Clip * true_child = 0;
  AVS_Clip * child = avs_take_clip(avs_array_elt(args, 0), env);
  AVS_Clip * clip = 0;

  if (!avs_is_field_based(avs_get_video_info(child))) {
    AVS_Value sclip = avs_new_value_clip(child);
    AVS_Value wargs = avs_new_value_array(&sclip, 1);
    AVS_Value rclip = avs_invoke(env, "SeparateFields", wargs, 0); 
    if (avs_is_error(rclip)) ERR(avs_as_error(rclip));
    true_child = child;
    child = avs_take_clip(rclip, env);
    avs_release_value(rclip);
    avs_release_value(wargs);
    avs_release_value(sclip);
  }

  child_v = avs_new_value_clip(child);
  clip = avs_new_c_filter(env, &fi, child_v, 1);
  avs_release_value(child_v); child_v = avs_void;
  fi->user_data = true_child;
  fi->get_frame = ibob_get_frame;
  fi->free_filter = ibob_free_filter;

  fi->vi.height *= 2;
  avs_clear_property(&fi->vi, AVS_IT_FIELDBASED);

  ret = avs_new_value_clip(clip);
  avs_release_clip(clip);
  return ret;
error:
  if (clip) avs_release_clip(clip); 
  return ret;
}

const char * avisynth_c_plugin_init(AVS_ScriptEnvironment * env)
{
  avs_add_function(env, "IBob", "c", ibob, 0);
  return "";
}

extern "C"
void bob_even(const byte * s, int s_pitch, byte * d, int d_pitch,
	      int row_size, int height)
{
  byte * d0 = d;
  byte * d1 = d0 + d_pitch;
  d_pitch *= 2;
  const byte * s0 = s; 
  const byte * s1 = s0 + s_pitch;
  for (int y = 0; y != height - 1; ++y) {
    for (int x = 0; x != row_size; ++x) {
      d0[x] = s0[x];
      d1[x] = (s0[x] + s1[x] + 1)/2;
    }
    d0 += d_pitch; d1 += d_pitch;
    s0 = s1; s1 += s_pitch;
  }
  for (int x = 0; x != row_size; ++x) {
    d0[x] = s0[x];
    d1[x] = s0[x];
  }
}

extern "C"
void bob_odd(const byte * s, int s_pitch, byte * d, int d_pitch,
	     int row_size, int height)
{
  byte * d0 = d;
  byte * d1 = d0 + d_pitch;
  d_pitch *= 2;
  const byte * s0 = 0; 
  const byte * s1 = s;
  for (int x = 0; x != row_size; ++x) {
    d0[x] = s1[x];
    d1[x] = s1[x];
  }
  for (int y = 1; y != height; ++y) {
    d0 += d_pitch; d1 += d_pitch;
    s0 = s1; s1 += s_pitch;
    for (int x = 0; x != row_size; ++x) {
      d0[x] = (s0[x] + s1[x] + 1)/2;
      d1[x] = s1[x];
    }
  }
}

extern "C"
void bob_chroma(const byte * s, int s_pitch, byte * d, int d_pitch,
		int row_size, int height)
{
  byte * d0 = d;
  byte * d1 = d0 + d_pitch;
  d_pitch *= 2;
  for (int y = 0; y != height; ++y) {
    for (int x = 0; x != row_size; ++x) {
      d0[x] = s[x];
      d1[x] = s[x];
    }
    d0 += d_pitch; d1 += d_pitch;
    s += s_pitch;
  }
}

extern "C"
void bob_even_isse_a(const byte * s, int s_pitch, byte * d, int d_pitch,
		     int row_size, int height)
{
  byte * d0 = d;
  byte * d1 = d0 + d_pitch;
  d_pitch *= 2;
  const byte * s0 = s; 
  const byte * s1 = s0 + s_pitch;
  for (int y = 0; y != height - 1; ++y) {
    for (int x = 0; x < row_size; x += 8) {
      asm("movq (%[s0],%0), %%mm0  \n\t"
	  "movq (%[s1],%0), %%mm1  \n\t"
	  "movq %%mm0, (%[d0],%0)  \n\t"
	  "pavgb %%mm0, %%mm1 \n\t"
	  "movq %%mm1, (%[d1],%0)  \n\t"
	  : : "r" (x), [s0]"r"(s0), [s1]"r"(s1), [d0]"r"(d0), [d1]"r"(d1)
	  : "memory");
    }
    d0 += d_pitch; d1 += d_pitch;
    s0 = s1; s1 += s_pitch;
  }
  for (int x = 0; x < row_size; x += 8) {
      asm("movq (%[s0],%0), %%mm0  \n\t"
	  "movq %%mm0, (%[d0],%0)  \n\t"
	  "movq %%mm0, (%[d1],%0)  \n\t"
	  : : "g" (x), [s0]"r"(s0), [d0]"r"(d0), [d1]"r"(d1)
	  : "memory");
  }
}

extern "C"
void bob_odd_isse_a(const byte * s, int s_pitch, byte * d, int d_pitch,
		    int row_size, int height)
{
  byte * d0 = d;
  byte * d1 = d0 + d_pitch;
  d_pitch *= 2;
  const byte * s0 = 0; 
  const byte * s1 = s;
  for (int x = 0; x < row_size; x += 8) {
    asm("movq (%[s1],%0), %%mm0  \n\t"
	"movq %%mm0, (%[d0],%0)  \n\t"
	"movq %%mm0, (%[d1],%0)  \n\t"
	: : "g" (x), [s1]"r"(s1), [d0]"r"(d0), [d1]"r"(d1)
	: "memory");
  }
  for (int y = 1; y != height; ++y) {
    d0 += d_pitch; d1 += d_pitch;
    s0 = s1; s1 += s_pitch;
    for (int x = 0; x < row_size; x += 8) {
      asm("movq (%[s1],%0), %%mm1  \n\t"
	  "movq (%[s0],%0), %%mm0  \n\t"
	  "movq %%mm1, (%[d1],%0)  \n\t"
	  "pavgb %%mm1, %%mm0 \n\t"
	  "movq %%mm0, (%[d0],%0)  \n\t"
	  : : "r" (x), [s0]"r"(s0), [s1]"r"(s1), [d0]"r"(d0), [d1]"r"(d1)
	  : "memory");
    }
  }
}


extern "C"
void bob_chroma_isse_a(const byte * s, int s_pitch, byte * d, int d_pitch,
		       int row_size, int height)
{
  byte * d0 = d;
  byte * d1 = d0 + d_pitch;
  d_pitch *= 2;
  for (int y = 0; y != height; ++y) {
    for (int x = 0; x < row_size; x += 8) {
      asm("movq (%[s],%0), %%mm0  \n\t"
	  "movq %%mm0, (%[d0],%0)  \n\t"
	  "movq %%mm0, (%[d1],%0)  \n\t"
	  : : "g" (x), [s]"r"(s), [d0]"r"(d0), [d1]"r"(d1)
	  : "memory");
    }
    d0 += d_pitch; d1 += d_pitch;
    s += s_pitch;
  }
}


AVS_VideoFrame * ibob_get_frame(AVS_FilterInfo * fi, int num)
{
  AVS_VideoFrame * src  = avs_get_frame(fi->child, num);
  bool parity = avs_get_parity(fi->child, num);
  AVS_VideoFrame * dest = avs_new_frame(fi->env, &fi->vi);
  
  if (parity) // even
    bob_even_isse(avs_get_read_ptr(src), avs_get_pitch(src),
	     avs_get_write_ptr(dest), avs_get_pitch(dest),
	     avs_get_row_size(src), avs_get_height(src));
  else
    bob_odd_isse(avs_get_read_ptr(src), avs_get_pitch(src),
	    avs_get_write_ptr(dest), avs_get_pitch(dest),
	    avs_get_row_size(src), avs_get_height(src));

  if (avs_is_yv12(&fi->vi)) {
    bob_chroma(avs_get_read_ptr_p(src, AVS_PLANAR_U), 
	       avs_get_pitch_p(src, AVS_PLANAR_U),
	       avs_get_write_ptr_p(dest, AVS_PLANAR_U), 
	       avs_get_pitch_p(dest, AVS_PLANAR_U),
	       avs_get_row_size_p(src, AVS_PLANAR_U), 
	       avs_get_height_p(src, AVS_PLANAR_U));
    bob_chroma(avs_get_read_ptr_p(src, AVS_PLANAR_V), 
	       avs_get_pitch_p(src, AVS_PLANAR_V),
	       avs_get_write_ptr_p(dest, AVS_PLANAR_V), 
	       avs_get_pitch_p(dest, AVS_PLANAR_V),
	       avs_get_row_size_p(src, AVS_PLANAR_V), 
	       avs_get_height_p(src, AVS_PLANAR_V));
  }
  avs_release_frame(src);
  return dest;
}

void ibob_free_filter(AVS_FilterInfo * fi)
{
  if (fi->user_data)
    avs_release_clip((AVS_Clip *)(fi->user_data));
}
